import z from 'zod';
import { ORIGIN_TYPE_PAYPAL, ORIGIN_TYPE_STRIPE } from '../origin/constants';
import { stripePaymentIntentIdParser } from '../stripe/parsers';
import { PAYIN_CHANNEL_API_LITERAL_TYPE, PAYIN_ORIGIN_TYPES, PAYIN_USER_API_LITERAL_TYPE } from './constants';
import { ChannelId } from '../channel/types';
import { cast } from '../utils/types';
import { UserId } from '../user/types';

declare type ReadonlyNonEmptyArray<A> = ReadonlyArray<A> & {
  readonly 0: A;
};

export const payinOriginTypeParser = z.enum(PAYIN_ORIGIN_TYPES);
export const payinOriginIdParser = stripePaymentIntentIdParser;
export const payinCompositeIdParser = z
  .object({
    provider: payinOriginTypeParser,
    originId: payinOriginIdParser,
  })
  .and(
    z.discriminatedUnion('provider', [
      z.object({
        provider: z.literal(ORIGIN_TYPE_STRIPE),
        originId: stripePaymentIntentIdParser,
      }),
      z.object({
        provider: z.literal(ORIGIN_TYPE_PAYPAL),
        originId: z.literal('paypal'), // TODO
      }),
    ])
  );
export const SEPARATOR = '::'; // just coincides with UUID_SEPARATOR
// "api src" - looks just like a normal graph src but is not graph src; abstractions are just similar
// (in truth, its because the older api already works like that but we'd like to both move from the graph terminology and keep compatibility)
// [8c2eab11-1e88-41dc-a520-ed22af399f96]
const makeApiIdParser = <
  L extends string,
  IdMap extends {
    [key in L]: Id;
  },
  Id extends string
>(
  literals: ReadonlyNonEmptyArray<L>
) => {
  const makeType = (literal: L) =>
    z.object({
      type: z.literal(literal),
      id: z.string().transform(cast<IdMap[L]>()),
    });
  // to procure type
  const first = makeType(literals[0]);

  return z
    .string()
    .min(SEPARATOR.length + 2 /*id and type*/)
    .transform((uuid, ctx) => {
      const [type, id] = uuid.split(SEPARATOR);
      const next = z.discriminatedUnion('type', [
        first,
        ...(literals.slice(1).map(makeType) as [typeof first, ...typeof first[]]),
      ]);
      const nextR = next.safeParse({ type, id: id as IdMap[L] });
      if (!nextR.success) {
        nextR.error.issues.forEach(ctx.addIssue.bind(ctx));
        return z.NEVER as z.infer<typeof next>;
      } else {
        return nextR.data;
      }
    });
};
const makeApiIdSerializedParser = <
  L extends string,
  IdMap extends {
    [key in L]: Id;
  },
  Id extends string
>(
  literals: ReadonlyNonEmptyArray<L>
) =>
  makeApiIdParser<L, IdMap, Id>(literals).transform(
    (cid) => `${(cid as unknown as { type: L }).type}${SEPARATOR}${(cid as unknown as { id: Id }).id}` as const
  );
const srcLiterals = [PAYIN_USER_API_LITERAL_TYPE] as const;
export const payinApiSrcSerializedParser = makeApiIdSerializedParser<
  typeof srcLiterals[number],
  {
    user: UserId;
  },
  UserId
>(srcLiterals);
export const payinApiSrcCompositeParser = makeApiIdParser<
  typeof srcLiterals[number],
  {
    user: UserId;
  },
  UserId
>(srcLiterals);
const dstLiterals = [PAYIN_CHANNEL_API_LITERAL_TYPE] as const;
export const payinApiDstSerializedParser = makeApiIdSerializedParser<
  typeof dstLiterals[number],
  {
    channel: ChannelId;
  },
  ChannelId
>(dstLiterals);
export const payinApiDstCompositeParser = makeApiIdParser<
  typeof dstLiterals[number],
  {
    channel: ChannelId;
  },
  ChannelId
>(dstLiterals);
