import z from 'zod';
import { SUBSCRIPTION_STATUS_INCOMPLETE, SUBSCRIPTION_STATUSES } from './constants';
import { StripePaymentIntentId, StripeSubscriptionId } from '../stripe/types';
import { Subset } from '../utils/types';
import { Currency } from '../currencies/types';
import {
  subscriptionCreateInputParser,
  subscriptionCreateResponseParser,
  subscriptionDeleteParamsParser,
  subscriptionListQueryParser,
  subscriptionListResponseParser,
  subscriptionUpdateInputParser,
  subscriptionUpdateParamsParser,
  subscriptionUpdateResponseParser,
} from './parsers';
import { PAYIN_ORIGIN_TYPES, PAYIN_PROVIDERS } from '../payin/constants';
import { ClientSecretMixin, PayinDst, PayinSrc } from '../payin/types';

export type SubscriptionId = StripeSubscriptionId;
export type SubscriptionProvider = typeof PAYIN_PROVIDERS[number];
export type SubscriptionStatus = typeof SUBSCRIPTION_STATUSES[number];
export type SubscriptionStatusIncomplete = typeof SUBSCRIPTION_STATUS_INCOMPLETE;
export type SubscriptionOriginType = typeof PAYIN_ORIGIN_TYPES[number];
type SubscriptionOriginIdPerTypeMapping = {
  paypal: never; // PaypalSubscriptionId,
  stripe: StripeSubscriptionId;
};
// eslint-disable-next-line @typescript-eslint/no-unused-vars
type AssertSubscriptionOriginIdPerTypeMapping1 = Subset<
  keyof SubscriptionOriginIdPerTypeMapping,
  SubscriptionOriginType
>;
// eslint-disable-next-line @typescript-eslint/no-unused-vars
type AssertSubscriptionOriginIdPerTypeMapping2 = Subset<
  SubscriptionOriginType,
  keyof SubscriptionOriginIdPerTypeMapping
>;
export type SubscriptionOriginIdForType<T extends SubscriptionOriginType> = SubscriptionOriginIdPerTypeMapping[T];

// amt, status, unit, maybe client_secret, maybe origin_id
// backward compat with current frontend
export type ApiSubscription<T extends SubscriptionOriginType = SubscriptionOriginType> = {
  amt: number;
  unit: Currency;
  src: PayinSrc;
  dst: PayinDst;
  originType: T;
  originId: SubscriptionOriginIdForType<T>;
  status: SubscriptionStatus;
};

export type ApiSubscriptionWithProceedContext<T extends SubscriptionOriginType = SubscriptionOriginType> =
  ApiSubscription<T> & ClientSecretMixin;

export type SubscriptionListApiQuery = z.input<typeof subscriptionListQueryParser>;
export type SubscriptionListApiQueryTransformed = z.output<typeof subscriptionListQueryParser>;
// eslint-disable-next-line @typescript-eslint/no-unused-vars
// type AssertSubscriptionListApiQuery = Subset<SubscriptionListApiQuery, PayinSrcDstApiQueryMixin>;
export type SubscriptionListApiResponse = z.output<typeof subscriptionListResponseParser>; // questionable whether we want to encode channel type there; probably not
export type SubscriptionUpdateApiParams = z.input<typeof subscriptionUpdateParamsParser>;
export type SubscriptionUpdateApiParamsTransformed = z.output<typeof subscriptionUpdateParamsParser>;
export type SubscriptionUpdateApiBody = z.input<typeof subscriptionUpdateInputParser>;
export type SubscriptionUpdateApiBodyTransformed = z.output<typeof subscriptionUpdateInputParser>;
export type SubscriptionUpdateApiResponse = z.infer<typeof subscriptionUpdateResponseParser>;
export type SubscriptionCreateApiBody = z.input<typeof subscriptionCreateInputParser>;
export type SubscriptionCreateApiBodyTransformed = z.output<typeof subscriptionCreateInputParser>;
export type SubscriptionCreateApiResponse = z.infer<typeof subscriptionCreateResponseParser>;
export type SubscriptionDeleteApiParams = z.input<typeof subscriptionDeleteParamsParser>;
export type SubscriptionDeleteApiParamsTransformed = z.output<typeof subscriptionDeleteParamsParser>;

// in a provider (at least stripe), a subscription is not a subscription yet; it can be considered as "subscription intent" up to some point
// we 'proceed' with paying for it as the next part of the flow
export type SubscriptionProviderProceedContext = {
  stripe: {
    // https://stripe.com/docs/billing/subscriptions/overview#how-payments-work-subscriptions
    paymentIntentId: StripePaymentIntentId;
    secret: string;
  };
  paypal: never; // TODO
};

export type CreateSubscriptionInProviderResult<OT extends SubscriptionOriginType> = {
  originId: SubscriptionOriginIdForType<OT>;
  subscription: ApiSubscription<OT>;
  proceedContext: SubscriptionProviderProceedContext[OT];
};
